/*
 * Copyright (C) 2012, Eric Wong <normalperson@yhbt.net>
 * License: GPLv3 or later (see COPYING for details)
 */
#include "cmogstored.h"

/*
 * we can lower this if we can test with lower values, NPTL minimum is 16K.
 * We also use syslog() and *printf() functions which take a lot of
 * stack under glibc, so we'll add BUFSIZ (8192 on glibc) to that
 */
#if MOG_LIBKQUEUE /* libkqueue uses quite a bit of stack */
#  define MOG_THR_STACK_SIZE (0)
#elif defined(__GLIBC__) || defined(__FreeBSD__)
#  define MOG_THR_STACK_SIZE ((16 * 1024) + MAX(8192,BUFSIZ))
#  if defined(PTHREAD_STACK_MIN) && (PTHREAD_STACK_MIN > MOG_THR_STACK_SIZE)
#    undef MOG_THR_STACK_SIZE
#    define MOG_THR_STACK_SIZE PTHREAD_STACK_MIN
#  endif
#else
#  define MOG_THR_STACK_SIZE (0)
#endif
static const size_t stacksize = (size_t)MOG_THR_STACK_SIZE;

void
mog_thrpool_start(struct mog_thrpool *tp, size_t n,
                  void *(*start_fn)(void *), void *arg)
{
	size_t i;
	pthread_t *thr;

	if (n == 0)
		n = 1;
	tp->threads = thr = xmalloc(sizeof(pthread_t) * n);
	tp->n_threads = n;

	for (i = 0; i < n; i++, thr++) {
		pthread_attr_t attr;

		CHECK(int, 0, pthread_attr_init(&attr));

		if (stacksize > 0) {
			CHECK(int, 0,
			      pthread_attr_setstacksize(&attr, stacksize));
		}

		CHECK(int, 0, pthread_create(thr, &attr, start_fn, arg));
		CHECK(int, 0, pthread_attr_destroy(&attr));
	}
}

void mog_thrpool_quit(struct mog_thrpool *tp, struct mog_queue *q)
{
	size_t i;

	for (i = 0; i < tp->n_threads; i++)
		CHECK(int, 0, pthread_cancel(tp->threads[i]));

	for (i = 0; i < tp->n_threads; i++) {
		pthread_t thr = tp->threads[i];
		int err;

		if (q) {
			/*
			 * if we can't rely on cancellation, keep poking
			 * the thread until it wakes up from cancellation.
			 */
			while ((err = pthread_kill(thr, SIGURG)) == 0)
				sched_yield();
			assert(err == ESRCH && "pthread_kill() usage bug");
		}

		CHECK(int, 0, pthread_join(thr, NULL));
	}

	mog_free_and_null(&tp->threads);
}
